;ARCMON.SYS
;Arcade Monitor Device Driver
;
;This is a fairly simple DOS Device Driver which allows VGA text mode and VGA mode 13h (320x200)
;to be displayed on an arcade monitor

;MACROS for setting VGA registers
MCRTC MACRO
	mov	ah,bl
	lodsb
	xchg	al,ah
	out	dx,ax
	inc	bl
	ENDM

MGEN	MACRO
	mov	ah,bl
	lodsb
	xchg	al,ah
	out	dx,ax
	ENDM

MATTR	MACRO
	mov	dx,3dah
	in		al,dx
	mov	dx,3c0h
	mov	al,bl
	or		al,32
	out	dx,al
	lodsb
	out	dx,al
	ENDM
;Code Segment (well - our one and only segment)
_TEXT	SEGMENT WORD PUBLIC 'CODE'
		ASSUME CS:_TEXT, DS:_TEXT, ES:NOTHING
;device driver header has to appear at start of segment
;so there's no PSP
	ORG     0h
;our device driver header
	dd  -1					;next driver in chain (set by DOS)
	dw  0C840h				;indicate we're a character based device
	dw	offset strategy		;our strategy proc
	dw  offset interrupt	;our interrupt proc
	db '$$ARCM$$'         	;our device name - 'Unusual' so as not to clash with other devices/filenames
;our variables
	rheader	dw 0,0
	cfgfile	db 'config.sys',0
;handle both upper and lowercase parameters
	xcent	db '/x '
	xcent1	db '/X '
	ycent	db '/y '
	ycent1	db '/Y '
	hval	db '/h '
	hval1	db '/H '
	vval	db '/v '
	vval1	db '/V '

	bread	dw ?
	blp		dw ?
	actx	db 8
	acty	db 8
	acth	db 105
	actv	db 4
	valof	dw ?
	fhandle	dw ?
	ioarea	db 2048 dup(0)
	numlen	db ?
	ddisp	db '$$$$$'
	topline	db 13,10
         	db 0c9h
         	db 35 dup(0cdh)
         	db 0bbh
         	db 24h
	midline	db 13,10
         	db 0c7h
         	db 35 dup(0c4h)
         	db 0b6h
         	db 24h
	botline	db 13,10
         	db 0c8h
         	db 35 dup(0cdh)
         	db 0bch,13,10
         	db 24h
;startup display banners
	banner	db 13,10,0bah,' Arcade Monitor Driver v1.0 Loaded ',0bah,24h
	info1	db 13,10,0bah,'     Horizontal Total      :       ',0bah,24h
	info2	db 13,10,0bah,'     Vertical Total        :       ',0bah,24h
	info3	db 13,10,0bah,'     Center X              :       ',0bah,24h
	info4	db 13,10,0bah,'     Center Y              :       ',0bah,24h


;register value sets for the arcade monitor modes supported
vgaregs13h:
;3c2
	db	231
;3d4 0 - 9
vgahtot db 105
	db 79,79,146
vgahrs db 85
vgahre db 31
vgavtot db 9
	db 17,0,64
;3d4 0ah - 0bh
	db 0,0
;3d4 10h - 17h
vgavrs	db	225
vgavre	db	164
	db	199
	db	40,64
	db	201,240,163
;Fancy a PC monitor 'scanlines' driver ?
;just change the first value in the following line from 9 to 1
;and the driver will produce a PC monitor 'scanline' mode
;3c4 1,4
	db   9,14

vgaregs3h:
;3c2
	db	231
;3d4 0 - 9
txthtot	db	105
	db	79,79,146
txthrs	db 85
txthre	db 31
txtvtot  db 9
	db	17,0,64+7
;3d4 0ah - 0bh
	db 5,7
;3d4 10h - 17h
txtvrs	db 225
txtvre	db 164
	db 199
	db 40,31
	db 219,240,163
;Fancy a PC monitor 'scanlines' driver ?
;just change the first value in the following line from 9 to 1
;and the driver will produce a PC monitor 'scanline' mode
;3c4 1,4
	db 9,2
;where we'll store the old BIOS vector
	BIOSSAVE DD      ?

strategy proc far
;this is the first thing called when the device driver is loaded
;at this point - DOS may not even exist, so we can't do that much
;just save the request header info into our var
	mov		cs:rheader[0],bx
	mov		cs:rheader[2],es
	ret
strategy endp


setmode proc
;This sets a given arcade monitor mode
	push	ax
	push	bx
	push	cx
	push	ds
	push	es
	push	si
	push	dx

	push	cs
	pop		ds
;clear VGA protection
	mov		dx,3d4h
	mov		al,11h
	out		dx,al
	inc		dx
	in 		al,dx
	and		al,127
	mov		ah,al
	dec		dx
	mov		al,11h
	out		dx,ax

	cld
	pop si
	push si

;3c2 misc. output registers
	lodsb
	mov		dx,3c2h
	out		dx,al
;3d4 the CRTC registers
	mov		dx,3d4h
	mov		cx,10
	xor		bx,bx
crlp1:
	MCRTC
	loop	crlp1

	mov		cx,2
	mov		bl,0ah
crlpx:
	MCRTC
	loop	crlpx
	mov		cx,8
	mov		bl,10h
crlp2:
	MCRTC
	loop	crlp2

;3c4 the sequencer registers
	mov		dx,3c4h
	mov		bl,1
	MGEN
	mov		bl,4
	MGEN

	pop		dx
	pop		si
	pop		es
	pop		ds
	pop		cx
	pop		bx
	pop		ax
	ret
setmode endp

; Start of device driver interrupt routine

interrupt proc far
	pushf
	push	ax
	push	bx
	push	cx
	push	dx
	push	ds
	push	es
	push	di
	push	si
	push	bp

	push	cs
	pop		ds
;interrupt routine is very simple
;if it's function 0 (Initialise) call init
;if it's not, just return 'OK'
	mov		si,rheader[0]
	mov		es,rheader[2]
	mov		bl,es:[si+2]
	cmp		bl,0
	jne		noinit
	call	init
noinit:
;always flag everything OK
	mov		ax,256
	mov		es:[si+3],ax

	pop		bp
	pop		si
	pop		di
	pop		es
	pop		ds
	pop		dx
	pop		cx
	pop		bx
	pop		ax
	popf

	ret

interrupt   ENDP



set8x8 proc
;function to set 8x8 font size
;needed if we're to fit on an arcade monitor
	push	ax
	push	bx
	push	es

	mov		al,02h
	mov		ah,11h
	xor		bx,bx
	mov		es,bx
	int		10h

	pop		es
	pop		bx
	pop		ax
	ret
set8x8 endp



newbios PROC
	push dx
;our new bios routine
;check for VGA mode request
	cmp		ah,0
	jne		callold
	push	ax

	and		al,127
;supported modes
	cmp		al,13h
	jne 	not13h
	mov		dx,offset vgaregs13h
	jmp		modesup
not13h:
	cmp		al,03h
	jne		notmode
	mov		dx,offset vgaregs3h
;set it to our arcade friendly mode
modesup:
	pop		ax
;supported arcade monitor mode
;call the original BIOS
	call	oldint10h
;set our arcade monitor mode
	call	setmode
;check if we need to reprogram the font
	cmp		dx,offset vgaregs3h
	jne		no8x8
	call	set8x8
no8x8:
	pop 	dx
	iret

notmode:
;unsupported mode - block and return
	pop		ax
	pop		dx
	iret

callold:
;any old bios call
	pop		dx
	call	oldint10h
	iret
newbios ENDP

oldint10h  PROC
;call the original BIOS
;pretend to be DOS and push the flags onto the stack before calling
	pushf
	call	dword ptr cs:BIOSSAVE
	ret
oldint10h  ENDP

;
;Everything from now on is discarded by DOS after the device driver is loaded
;

init proc near
	push	bx
	push	dx
	push	ds

	push	cs
	pop		ds
;At this point - we can assume some of DOS is there
;so we can get the command line arguments
;As device drivers have no PSP (where the command line arguments would normally be placed by DOS)
;we need to read in CONFIG.SYS to see if there are any arguments on the line
	mov		ah,3dh
	mov		al,0
	lea		dx,cfgfile
	int		21h
   	jc		ferror
	mov		fhandle,ax
;we managed to open config.sys, so read in the values
	call	readcfg
;close config.sys
	mov		bx,fhandle
	mov		ah,3eh
	int		21h
ferror:
;pass the arguments/defaults on
	call	centermode
;then set an arcade monitor mode
	push	dx
	mov		dx,offset vgaregs3h
	call	setmode
	pop		dx
;set text to 8x8
	call	set8x8
;show our banner
	call	showbanner
;setup our memory requirements so DOS knows how big we are
;- everything up to the start of this function will remain in memory
;after the device driver is loaded
	mov		ax,offset init
	mov		es:[si+0eh],ax
	mov		es:[SI+10h],cs

	mov		ax,0

	call	repbios
	pop		ds
	pop		dx
	pop		bx
	ret
init endp


readcfg proc
;read in config.sys
	push	ax
   	push	bx
   	push	cx
   	push	dx

	mov		ah,3fh
	mov		bx,fhandle
	lea		dx,ioarea
	mov		cx,2048
	int		21h

	jc		rerr
	mov		bread,ax

;get arguments
	lea		bx,ycent
	lea		dx,acty
	call	getcfgval
	lea		bx,ycent1
	call	getcfgval

	lea		bx,xcent
	lea		dx,actx
	call	getcfgval
	lea		bx,xcent1
	call	getcfgval


	lea		bx,hval
	lea		dx,acth
	call	getcfgval
	lea		bx,hval1
	call	getcfgval

	lea		bx,vval
	lea		dx,actv
	call	getcfgval
	lea		bx,vval1
	call	getcfgval

rerr:

	pop		dx
	pop		cx
	pop		bx
	pop		ax

	ret
readcfg endp

getcfgval proc
;get a config value from the given line
	push	ax
	push	bx
	push	cx
	push	dx
	push	di
   	push	si

	mov		valof,dx
	xor		cx,cx
cfg1:
	push	cx
;get the io buffer
	lea		si,ioarea
;and the string to find
	mov		di,bx
;skip to the current position
	add		si,cx
;see if the string matches
	mov		cx,3
	call	cmpstr
	cmp		ax,0
	jne		ncfgm
;found it
	pop		cx

	lea		dx,ioarea
	add		dx,cx
;so get the number
	call	getnum

;jump to end of function
	jmp		cfgfin
ncfgm:
	pop		cx
	inc		cx
	cmp		cx,bread
	jne		cfg1
cfgfin:

	pop		si
	pop		di
	pop		dx
	pop		cx
	pop		bx
	pop		ax
	ret
getcfgval endp

getnum   proc
;get the number after the config item as a string
	push	ax
	push	di
   	push	si

	mov		byte ptr numlen,0
	mov		di,dx
getn1:
	cmp		byte ptr [di],32
	je		getn2
	inc		di
	jmp		getn1
getn2:
	cmp		byte ptr [di],32
	jne		getn3
	inc		di
	jmp		getn2
getn3:
	lea		si,ddisp
getn4:
	mov		al,byte ptr [di]
	cmp		al,32
	jle		getn5
	mov		byte ptr [si],al
	inc		byte ptr numlen
	cmp		numlen,byte ptr 3
	je		getn5
	inc		si
	inc		di
	jmp		getn4
getn5:
	inc		si
	mov		byte ptr [si],24h

	call	conval

	pop		si
	pop		di
	pop		ax
	ret
getnum   endp

conval proc
;convert a string into a number
	push	ax
	push	bx
	push	cx
	push	dx
	push	di

	mov		bx,0
	lea		di,ddisp
	mov		ch,0
	mov		cl,numlen
con1:
	mov		al,[di]

	cmp		cx,3
	jne		con2
	sub		al,'0'
	cmp		al,9
	jg		con4
	mov		ah,0
	mov		dl,100
	mul		dl
	mov		bx,ax
	jmp		con4
con2:
	cmp		cx,2
	jne		con3
	sub		al,'0'
	cmp		al,9
	jg		con4
	mov		ah,0
	mov		dl,10
	mul		dl
	add		bx,ax
	jmp		con4
con3:
	sub		al,'0'
	cmp		al,9
	jg		con4
	mov		ah,0
	add		bx,ax
con4:
	inc		di
	loop	con1

	mov		ax,bx
	mov		di,valof
	mov		[di],al


	pop		di
	pop		dx
	pop		cx
	pop		bx
	pop		ax
	ret
conval endp

dispval proc
;convert a number into a string
	push	ax
	push	bx
	push	cx
	push	dx

	mov		bl,al
	mov		bh,0
	cmp		bx,100
	jl		no100
	mov		dl,'0'
d100lp:
	cmp		bx,100
	jl		d100fin
	sub		bl,100
	inc		dl
	jmp		d100lp
d100fin:
	mov		ah,06h
	int		21h
	jmp		r10s
no100:
	cmp		bx,10
	jl		no10
r10s:
	mov		dl,'0'
d10lp:
	cmp		bx,10
	jl		d10fin
	sub		bl,10
	inc		dl
	jmp		d10lp
d10fin:
	mov		ah,06h
	int		21h
no10:
	mov		dl,'0'
	add		dl,bl
	mov		ah,06h
	int		21h

	pop		dx
	pop		cx
	pop		bx
	pop		ax
	ret
dispval endp

cmpstr proc
;compare one string with another
	push	bx
	mov		ax,0
cmp1:
	mov		bl,[si]
	cmp		bl,[di]
	jne		badend
	inc		si
	inc		di
	loop	cmp1
	pop		bx
	ret
badend:
	mov		ax,1
	pop		bx
	ret
cmpstr endp

centermode proc
;store our configurable values
	push	ax
	push	bx
	mov		al,acth
	mov		vgahtot,al
	mov		txthtot,al
;get hrs
	mov		al,vgahrs
;get the center value
	mov		bl,actx
	sub		bl,8
	add		al,bl
	mov		vgahrs,al
	mov		txthrs,al
	add		al,10
	and		al,31
	mov		vgahre,al
	mov		txthre,al
;get vrs
	mov		al,vgavrs
;get the center value
	mov		bl,acty
	sub		bl,8
	add		al,bl
	mov		vgavrs,al
	mov		txtvrs,al
	add		al,2
	and		al,15
	mov		bl,vgavre
	and		bl,240
	or		bl,al
	mov		vgavre,bl
	mov		txtvre,bl

	pop		bx
	pop		ax

	ret
centermode ENDP

showbanner proc
;display startup banner
	push	dx
	push	ds

	push	cs
	pop 	ds
	mov 	ah,9

	mov 	dx,offset topline
	int 	21h
	mov		dx,offset banner
	int		21h
	mov		dx,offset midline
	int		21h

	mov		dx,offset info1
	int		21h

	mov		al,acth
	call	backcur
	call	dispval

	mov		dx,offset info2
	int		21h

	mov		al,actv
	call	backcur
	call	dispval

	mov		dx,offset info3
	int		21h

	mov		al,actx
	call	backcur
	call	dispval

	mov		dx,offset info4
	int		21h

	mov		al,acty
	call	backcur
	call	dispval

	mov		dx,offset botline
	int		21h

	pop		ds
	pop		dx
	ret
showbanner endp

repbios proc
;replace existing BIOS vector with our own
	push	ax
	push	dx
	push	ds
	push	es
;be on the safe side and use DOS
;get the existing vector
	mov		al,10h
	mov		ah,35h
	int		21h
	mov		cs:word ptr BIOSSAVE,bx
	mov		cs:word ptr BIOSSAVE+2,es
;replace with out new vector
	push	ds
	mov   	ax,cs
	mov		ds,ax
	mov		dx,offset newbios
	mov   	al,10h
	mov   	ah,25h
	int   	21h
	pop   	ds

	pop		es
	pop		ds
	pop		dx
	pop		ax
	ret
repbios endp

backcur proc
;move the cursor back 7 places
;used by the display banner when the device driver is loaded
	push	ax
	push	bx
	push	dx

	mov		ah,3
	mov		bh,0
	int		10h

	mov		ah,2
	sub		dl,7

	int		10h

	pop		dx
	pop		bx
	pop		ax
	ret
backcur endp

_TEXT       ends
			end
